import torch
import torch.distributed as dist

from torch import Tensor

from abc import *
from typing import *


class Work:
    def __init__(self) -> None:
        pass
    
    @abstractmethod
    def wait(self) -> None:
        raise NotImplementedError


class NcclAllToAllWork(Work):
    def __init__(self,
        work: Any,
    ) -> None:
        super().__init__()
        self._work = work
    
    def wait(self) -> None:
        self._work.wait()


class GlooAllToAllWork(Work):
    def __init__(self,
        recv_works: List[Any],
        send_works: List[Any],
        output_tensor_list: List[Tensor],
        buffer_tensor_list: List[Tensor],
    ) -> None:
        super().__init__()
        self.recv_works = recv_works
        self.send_works = send_works
        self.output_tensor_list = output_tensor_list
        self.buffer_tensor_list = buffer_tensor_list
    
    def wait(self) -> None:
        rank = dist.get_rank()
        world_size = dist.get_world_size()
        
        for i in range(1, world_size):
            recv_i = (rank - i + world_size) % world_size
            self.recv_works[recv_i].wait()
            
            output_device = self.output_tensor_list[recv_i].device
            self.output_tensor_list[recv_i][:] = self.buffer_tensor_list[recv_i].to(output_device)
        
        for i in range(1, world_size):
            send_i = (rank + i) % world_size
            self.send_works[send_i].wait()


def all_to_all(
    output_tensor_list: List[Tensor],
    input_tensor_list: List[Tensor],
    async_op: bool = False,
) -> Optional[Work]:
    assert len(output_tensor_list) == len(input_tensor_list)
    
    backend = dist.get_backend()
    if backend == "nccl":
        work = dist.all_to_all(
            output_tensor_list=output_tensor_list,
            input_tensor_list=input_tensor_list,
            async_op=True,
        )
        work = NcclAllToAllWork(work)
        return work if async_op else work.wait()
    elif backend == "gloo":
        rank = dist.get_rank()
        world_size = dist.get_world_size()
        
        input_tensor_list = list(input_tensor_list)
        buffer_tensor_list = [None,] * world_size
        for i in range(1, world_size):
            k = (rank + i) % world_size
            input_tensor_list[k] = input_tensor_list[k].cpu()
            buffer_tensor_list[k] = torch.empty_like(output_tensor_list[k], device="cpu")
            
        output_tensor_list[rank][:] = input_tensor_list[rank]
        
        send_works = [None,] * world_size
        recv_works = [None,] * world_size
        for i in range(1, world_size):
            send_i = (rank + i) % world_size
            recv_i = (rank - i + world_size) % world_size
            
            send_works[send_i] = dist.isend(input_tensor_list[send_i], dst=send_i)
            recv_works[recv_i] = dist.irecv(buffer_tensor_list[recv_i], src=recv_i)
        
        work = GlooAllToAllWork(recv_works, send_works, output_tensor_list, buffer_tensor_list)
        return work if async_op else work.wait()
    else:
        raise ValueError(f"unsupported backend: {backend}") 


def all_to_all_v(
    output_tensor_list: List[Tensor],
    input_tensor_list: List[Tensor],
    async_op: bool = False,
) -> Optional[Work]:
    assert len(output_tensor_list) == len(input_tensor_list)
    
    input_tensor_sizes = []
    for x in input_tensor_list:
        assert x.dim() == 1
        input_tensor_sizes.append(x.numel())
    
    input_tensor_sizes = torch.tensor(
        input_tensor_sizes, dtype=torch.long, device=input_tensor_list[0].device)
    output_tensor_sizes = torch.empty_like(input_tensor_sizes)
    
    dist.all_to_all_single(
        output=output_tensor_sizes,
        input=input_tensor_sizes,
    )
    
    for i, sz in enumerate(output_tensor_sizes.tolist()):
        output_tensor_list[i] = torch.empty(
            sz, dtype=input_tensor_list[i].dtype, device=input_tensor_list[i].device)
        
    return all_to_all(
        output_tensor_list=output_tensor_list,
        input_tensor_list=input_tensor_list,
        async_op=async_op,
    )

