import torch
import torch.distributed as dist

from torch import Tensor, LongTensor
from torch_sparse import SparseTensor

from core.distributed.a2a import all_to_all_v

from torch.types import _device
from typing import *



class P2PIndex:
    def __init__(self,
        local_node_table: LongTensor,
        local_edge_table: LongTensor,
        local_edge_route: LongTensor,
        device: Union[_device, str, None] = None,
    ) -> None:
        self._init_device(device)
        self._init_node_map(local_node_table)
        self._init_edge_map(local_edge_table, local_edge_route)
    
    def _init_device(self, device: Union[_device, str, None]) -> None:
        if device is None:
            self._device = torch.cuda.current_device()
        else:
            self._device = torch.device(device)
    
    def _init_node_map(self,
        local_node_table: LongTensor,
    ) -> None:
        assert local_node_table.dim() == 1
        
        ind2vid = local_node_table.to(self._device)
        vid2ind = SparseTensor.from_edge_index(
            edge_index=torch.vstack([torch.zeros_like(ind2vid), ind2vid]),
            edge_attr=torch.arange(ind2vid.size(0)).type_as(ind2vid)
        ).coalesce("min")
        
        self._nbuf_ind2vid: LongTensor = ind2vid
        self._nbuf_vid2ind: SparseTensor = vid2ind
        
    def _init_edge_map(self,
        local_edge_table: LongTensor,
        local_edge_route: LongTensor,
    ) -> None:
        self._init_edge_buf(local_edge_table)
        self._local_edge_index: LongTensor = self.ebuf_vid2ind(local_edge_table)
        
        assert local_edge_route.dim() == 2
        assert local_edge_route.size(0) == 2
        
        vid, rot = local_edge_route.to(self._device)
        vid2rot = SparseTensor.from_edge_index(
            edge_index=torch.vstack([torch.zeros_like(vid), vid]),
            edge_attr=rot,
        ).coalesce("min")
        _, vid, rot = vid2rot.csr()
        
        assert vid.size(0) == self.ebuf.size(0)
        
        ind = self.ebuf_vid2ind(vid)
        ind, perm = ind.sort()
        ind2rot = rot[perm]
        
        self._ebuf_ind2rot: LongTensor = ind2rot
        self._ebuf_vid2rot: SparseTensor = vid2rot
        
    def _init_edge_buf(self,
        local_edge_table: LongTensor,
    ) -> None:
        assert local_edge_table.dim() == 2
        assert local_edge_table.size(0) == 2
        
        ind2vid = local_edge_table.to(self._device)
        
        ind0 = torch.vstack([torch.zeros_like(ind2vid[0]), ind2vid[0]])
        ind1 = torch.vstack([torch.zeros_like(ind2vid[1]), ind2vid[1]])
        inds = torch.hstack([ind0, ind1])
        
        val0 = torch.zeros_like(ind2vid[0])
        val1 = torch.ones_like(ind2vid[1])
        vals = torch.hstack([val0, val1])
        
        _, col, tag = SparseTensor.from_edge_index(inds, vals).coalesce("max").csr()
        col0 = col[tag == 0]
        col1 = col[tag == 1]
        ind2vid = torch.hstack([col1, col0])
        vid2ind = SparseTensor.from_edge_index(
            edge_index=torch.vstack([torch.zeros_like(ind2vid), ind2vid]),
            edge_attr=torch.arange(ind2vid.size(0)).type_as(ind2vid)
        ).coalesce("min")
        
        self._ebuf_sp_size: int = col1.size(0)
        self._ebuf_ind2vid: LongTensor = ind2vid
        self._ebuf_vid2ind: SparseTensor = vid2ind
    
    @property
    def local_edge_index(self) -> LongTensor:
        return self._local_edge_index
    
    @property
    def nbuf(self) -> LongTensor:
        return self._nbuf_ind2vid
    
    @property
    def ebuf(self) -> LongTensor:
        return self._ebuf_ind2vid
    
    @property
    def ebuf_sp_size(self) -> int:
        return self._ebuf_sp_size
    
    def nbuf_ind2vid(self, ind: LongTensor) -> LongTensor:
        return self._nbuf_ind2vid[ind.view(-1)].view(*ind.size())
    
    def nbuf_vid2ind(self, vid: LongTensor) -> LongTensor:
        return self._nbuf_vid2ind[0, vid.view(-1)].csr()[2].view(*vid.size())
    
    def ebuf_ind2vid(self, ind: LongTensor) -> LongTensor:
        return self._ebuf_ind2vid[ind.view(-1)].view(*ind.size())
    
    def ebuf_vid2ind(self, vid: LongTensor) -> LongTensor:
        return self._ebuf_vid2ind[0, vid.view(-1)].csr()[2].view(*vid.size())
    
    def ebuf_ind2rot(self, ind: LongTensor) -> LongTensor:
        return self._ebuf_ind2rot[ind.view(-1)].view(*ind.size())
        
    def ebuf_vid2rot(self, vid: LongTensor) -> LongTensor:
        return self._ebuf_vid2rot[0, vid.view(-1)].csr()[2].view(*vid.size())
    
    def gather_index(self, src: bool, dst: bool) -> Tuple[int, List[LongTensor], List[LongTensor]]:
        assert src or dst
        world_size = dist.get_world_size()
        
        dind = []
        if src: dind.append(self._local_edge_index[0])
        if dst: dind.append(self._local_edge_index[1])
        dind: LongTensor = torch.hstack(dind).unique()
        
        dvid = self.ebuf_ind2vid(dind)
        drot = self.ebuf_ind2rot(dind)
        
        recv_ind = []
        recv_vid = []
        for i in range(world_size):
            mask = (drot == i)
            recv_ind.append(dind[mask])
            recv_vid.append(dvid[mask])
            
        send_vid = [None] * len(recv_vid)
        all_to_all_v(send_vid, recv_vid)
        
        send_ind = []
        for i in range(world_size):
            send_ind.append(self.nbuf_vid2ind(send_vid[i]))
        
        dst_size = dind.max().item() + 1
        return dst_size, send_ind, recv_ind
    
    def scatter_index(self, src: bool, dst: bool) -> Tuple[int, List[LongTensor], List[LongTensor]]:
        assert src or dst
        world_size = dist.get_world_size()
        
        sind = []
        if src: sind.append(self._local_edge_index[0])
        if dst: sind.append(self._local_edge_index[1])
        sind: LongTensor = torch.hstack(sind).unique()
        
        svid = self.ebuf_ind2vid(sind)
        srot = self.ebuf_ind2rot(sind)
        
        send_ind = []
        send_vid = []
        for i in range(world_size):
            mask = (srot == i)
            send_ind.append(sind[mask])
            send_vid.append(svid[mask])
        
        recv_vid = [None] * len(send_vid)
        all_to_all_v(recv_vid, send_vid)
        
        recv_ind = []
        for i in range(world_size):
            recv_ind.append(self.nbuf_vid2ind(recv_vid[i]))
        
        src_size = sind.max().item() + 1
        return src_size, send_ind, recv_ind
    
    @property
    def device(self) -> _device:
        return self._device
    
    def to(self, device: Union[_device, str]):
        self._device = torch.device(device)
        for key in dir(self):
            val = getattr(self, key)
            if isinstance(val, Tensor):
                setattr(self, key, val.to(self._device))
            elif isinstance(val, SparseTensor):
                setattr(self, key, val.to(self._device))
        return self
    
    def cuda(self):
        device = torch.cuda.current_device()
        return self.to(device)
    
    def cpu(self):
        device = torch.device("cpu")
        return self.to(device)
        
        