import torch
import torch.autograd as autograd

from torch import Tensor, LongTensor
from typing import *

from torch_scatter import scatter
from core.distributed.a2a import all_to_all, all_to_all_v


class ReducerOptions:
    def __init__(self,
        src_size: int,
        dst_size: int,
        send_inds: List[LongTensor],
        recv_inds: List[LongTensor],
        reduce_op: str,
    ) -> None:
        self.src_size = src_size
        self.dst_size = dst_size
        
        self.send_inds = send_inds
        self.recv_inds = recv_inds
        
        self.reduce_op = reduce_op
    
    def backward(self):
        return ReducerOptions(
            src_size=self.dst_size,
            dst_size=self.src_size,
            send_inds=self.recv_inds,
            recv_inds=self.send_inds,
            reduce_op="sum",
        )

class Reducer(autograd.Function):
    @staticmethod
    def forward(
        ctx: autograd.function.FunctionCtx,
        src: Tensor,
        opt: ReducerOptions,
    ) -> Tensor:
        assert src.size(0) == opt.src_size
        
        send_data = []
        for ind in opt.send_inds:
            send_data.append(src[ind])
            
        recv_data = []
        for ind in opt.recv_inds:
            recv_data.append(torch.empty(ind.size(0), *src.shape[1:], dtype=src.dtype, device=src.device))
        
        all_to_all(recv_data, send_data)
        
        recv_inds = torch.cat(opt.recv_inds, dim=0)
        recv_data = torch.cat(recv_data, dim=0)
        dst = scatter(recv_data, recv_inds, dim=0, dim_size=opt.dst_size)
            
        ctx.saved_opt = opt.backward()
        return dst
    
    @staticmethod
    def backward(
        ctx: autograd.function.FunctionCtx,
        grad_output: Tensor,
    ):
        return Reducer.apply(grad_output, ctx.saved_opt), None


class DiffReducerOptions:
    def __init__(self,
        acti_src_memory: Tensor,
        acti_dst_memory: Tensor,
        grad_src_memory: Tensor,
        grad_dst_memory: Tensor,
        send_inds: List[LongTensor],
        recv_inds: List[LongTensor],
        reduce_op: str,
    ) -> None:
        
        self.acti_src_memory = acti_src_memory
        self.acti_dst_memory = acti_dst_memory
        
        self.grad_src_memory = grad_src_memory
        self.grad_dst_memory = grad_dst_memory
        
        self.send_inds = send_inds
        self.recv_inds = recv_inds
        
        self.reduce_op = reduce_op
    
    def backward(self):
        return DiffReducerOptions(
            acti_src_memory=self.grad_dst_memory,
            acti_dst_memory=self.grad_src_memory,
            grad_src_memory=self.acti_dst_memory,
            grad_dst_memory=self.acti_src_memory,
            send_inds=self.recv_inds,
            recv_inds=self.send_inds,
            reduce_op="sum",
        )


class DiffReducer(autograd.Function):
    @staticmethod
    def forward(
        ctx: autograd.function.FunctionCtx,
        src: Tensor,
        opt: DiffReducerOptions,
    ) -> Tensor:
        assert src.size() == opt.acti_src_memory.size()
        
        diff_src = src - opt.acti_src_memory
        diff_src = diff_src.type(torch.float16)
        
        send_data = []
        for ind in opt.send_inds:
            send_data.append(src[ind])
            
        recv_data = [None] * len(opt.recv_inds)
        all_to_all_v(recv_data, send_data)
        
        recv_inds = torch.cat(opt.recv_inds, dim=0)
        recv_data = torch.cat(recv_data, dim=0)
        
        recv_data = recv_data.type_as(opt.acti_dst_memory) + opt.acti_dst_memory[recv_inds]
        dst = scatter(recv_data, recv_inds, dim=0, dim_size=opt.acti_dst_memory.size(0))
            
        ctx.saved_opt = opt.backward()
        return dst
    
    @staticmethod
    def backward(
        ctx: autograd.function.FunctionCtx,
        grad_output: Tensor,
    ):
        return DiffReducer.apply(grad_output, ctx.saved_opt), None