import torch
import torch.distributed as dist

from torch_sparse import SparseTensor
from torch_scatter import scatter

from torch import Tensor, LongTensor
from torch.types import _device
from typing import *

from core.distributed.ind import P2PIndex
import core.distributed.ops as ops

import contextlib


class PGraph:
    def __init__(self,
        local_node_table: LongTensor,
        local_edge_table: LongTensor,
        local_edge_route: LongTensor,
        device: Union[_device, str, None] = None,
    ) -> None:
        self._p2p_index = P2PIndex(
            local_node_table=local_node_table,
            local_edge_table=local_edge_table,
            local_edge_route=local_edge_route,
            device=device
        )
        self._device = self._p2p_index.device
        self._nbuf_size = self._p2p_index.nbuf.size(0)
        self._ebuf_size = self._p2p_index.ebuf.size(0)
        self._local_edge_index = self._p2p_index.local_edge_index#.clone()
    
    def _gather_index_names(self, src: bool, dst: bool) -> Tuple[str, str, str]:
        tag = f"{int(src)}{int(dst)}"
        gbuf_size = f"_gather_{tag}_gbuf_size"
        send_inds = f"_gather_{tag}_send_inds"
        recv_inds = f"_gather_{tag}_recv_inds"
        return gbuf_size, send_inds, recv_inds
    
    def _scatter_index_names(self, src: bool, dst: bool) -> Tuple[str, str, str]:
        tag = f"{int(src)}{int(dst)}"
        sbuf_size = f"_scatter_{tag}_sbuf_size"
        send_inds = f"_scatter_{tag}_send_inds"
        recv_inds = f"_scatter_{tag}_recv_inds"
        return sbuf_size, send_inds, recv_inds
    
    def _gather_index(self, src: bool, dst: bool) -> Tuple[int, List[LongTensor], List[LongTensor]]:
        gbuf_size_name, send_inds_name, recv_inds_name = self._gather_index_names(src, dst)
        gbuf_size = getattr(self, gbuf_size_name)
        send_inds = getattr(self, send_inds_name)
        recv_inds = getattr(self, recv_inds_name)
        return gbuf_size, send_inds, recv_inds
    
    def _scatter_index(self, src: bool, dst: bool) -> Tuple[int, List[LongTensor], List[LongTensor]]:
        sbuf_size_name, send_inds_name, recv_inds_name = self._scatter_index_names(src, dst)
        sbuf_size = getattr(self, sbuf_size_name)
        send_inds = getattr(self, send_inds_name)
        recv_inds = getattr(self, recv_inds_name)
        return sbuf_size, send_inds, recv_inds
    
    def _buf_gather(self,
        x: Tensor,
        src: bool,
        dst: bool,
        reduce_op: str,
        cache: bool,
    ) -> Tensor:
        gbuf_size, send_inds, recv_inds = self._gather_index(src, dst)
        assert x.size(0) == self._nbuf_size
        assert reduce_op in {"sum", "max", "min"}
        
        opt = ops.ReducerOptions(
            src_size=x.size(0),
            dst_size=gbuf_size,
            send_inds=send_inds,
            recv_inds=recv_inds,
            reduce_op=reduce_op,
        )
        return ops.Reducer.apply(x, opt)
    
    def _buf_scatter(self,
        x: Tensor,
        src: bool,
        dst: bool,
        reduce_op: str,
        cache: bool,
    ) -> Tensor:
        sbuf_size, send_inds, recv_inds = self._scatter_index(src, dst)
        assert x.size(0) == sbuf_size
        assert reduce_op in {"sum", "max", "min"}
        
        opt = ops.ReducerOptions(
            src_size=x.size(0),
            dst_size=self._nbuf_size,
            send_inds=send_inds,
            recv_inds=recv_inds,
            reduce_op=reduce_op,
        )
        return ops.Reducer.apply(x, opt)
    
    def enable_gather(self, src: bool = False, dst: bool = False):
        if not src and not dst:
            src = True
            
        gbuf_size_name, send_inds_name, recv_inds_name = self._gather_index_names(src, dst)
        if hasattr(self, gbuf_size_name):
            return self
        
        gbuf_size, send_inds, recv_inds = self._p2p_index.gather_index(src, dst)
        send_inds = [x.to(self._device) for x in send_inds]
        recv_inds = [x.to(self._device) for x in recv_inds]
        
        setattr(self, gbuf_size_name, gbuf_size)
        setattr(self, send_inds_name, send_inds)
        setattr(self, recv_inds_name, recv_inds)
        return self
    
    def enable_scatter(self, src: bool = False, dst: bool = False):
        if not src and not dst:
            dst = True
            
        sbuf_size_name, send_inds_name, recv_inds_name = self._scatter_index_names(src, dst)
        if hasattr(self, sbuf_size_name):
            return self
        
        sbuf_size, send_inds, recv_inds = self._p2p_index.scatter_index(src, dst)
        send_inds = [x.to(self._device) for x in send_inds]
        recv_inds = [x.to(self._device) for x in recv_inds]
        
        setattr(self, sbuf_size_name, sbuf_size)
        setattr(self, send_inds_name, send_inds)
        setattr(self, recv_inds_name, recv_inds)
        return self
    
    def enable_all_gather(self):
        self.enable_gather(src=True, dst=True)
        self.enable_gather(src=True, dst=False)
        self.enable_gather(src=False, dst=True)
        return self
    
    def enable_all_scatter(self):
        self.enable_scatter(src=True, dst=True)
        self.enable_scatter(src=True, dst=False)
        self.enable_scatter(src=False, dst=True)
        return self
            
    def gather(self,
            x: Tensor,
            src: bool = True,
            dst: bool = False,
            cache: bool = False,
        ) -> Union[Tensor, Tuple[Tensor, Tensor]]:
        gbuf = self._buf_gather(x, src, dst, reduce_op="sum", cache=cache)
        
        if src and dst:
            m_j = gbuf[self._local_edge_index[0]]
            m_i = gbuf[self._local_edge_index[1]]
            return m_i, m_j
        elif src and not dst:
            m_j = gbuf[self._local_edge_index[0]]
            return m_j
        elif not src and dst:
            m_i = gbuf[self._local_edge_index[1]]
            return m_i
        else:
            raise RuntimeError("not src and not dst")
            
    def scatter(self,
        x: Tensor,
        src: bool = False,
        dst: bool = True,
        reduce_op: str = "mean",
        cache: bool = False,
        eps: float = 1e-12,
    ) -> Tensor:
        if src and dst:
            ind = self._local_edge_index.view(-1)
        elif src and not dst:
            ind = self._local_edge_index[0]
        elif not src and dst:
            ind = self._local_edge_index[1]
        else:
            raise RuntimeError("not src and not dst")
        
        if reduce_op == "mean":
            w = torch.ones(x.size(0), 1, dtype=x.dtype, device=x.device)
            w = scatter(w, ind, dim=0, reduce="sum")
            w = self._buf_scatter(w, src, dst, reduce_op="sum", cache=cache)
            
            x = scatter(x, ind, dim=0, reduce="sum")
            x = self._buf_scatter(x, src, dst, reduce_op="sum", cache=cache)
            return x / (w + eps)
        else:
            x = scatter(x, ind, dim=0, reduce=reduce_op)
            return self._buf_scatter(x, src, dst, reduce_op=reduce_op, cache=cache)
            
    def softmax(self,
        a: Tensor,
        src: bool = False,
        dst: bool = True,
        eps: float = 1e-12,
    ):
        assert src or dst
        
        with torch.no_grad():
            m = self.scatter(a, src, dst, reduce_op="max")
            if src and dst:
                m_i, m_j = self.gather(m, src, dst)
                m = m_i + m_j
            else:
                m = self.gather(m, src, dst)
        a = (a - m).exp()
        
        s = self.scatter(a, src, dst, reduce_op="sum")
        if src and dst:
            s_i, s_j = self.gather(s, src, dst)
            s = s_i + s_j
        else:
            s = self.gather(s, src, dst)
        return a / (s + eps)
        
    @property
    def device(self) -> _device:
        return self._device
    
    def to(self, device: Union[_device, str]):
        self._device = torch.device(device)
        for key in dir(self):
            val = getattr(self, key)
            if isinstance(val, (Tensor, SparseTensor)):
                setattr(self, key, val.to(self._device))
            elif isinstance(val, (list, tuple)):
                new_val = []
                for x in val:
                    if isinstance(x, (Tensor, SparseTensor)):
                        x = x.to(self._device)
                    new_val.append(x)
                if isinstance(val, tuple):
                    new_val = tuple(new_val)
                setattr(self, key, new_val)
        return self
    
    def cuda(self):
        device = torch.cuda.current_device()
        return self.to(device)
    
    def cpu(self):
        device = torch.device("cpu")
        return self.to(device)