import torch
import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor
from typing import *

import copy
import inspect

from core.distributed import PGraph
from torch_geometric.nn import JumpingKnowledge
from torch_geometric.nn.resolver import activation_resolver, normalization_resolver


class BasicGNN(nn.Module):
    r"""An abstract class for implementing basic GNN models.

    Args:
        in_channels (int or tuple): Size of each input sample, or :obj:`-1` to
            derive the size from the first input(s) to the forward method.
            A tuple corresponds to the sizes of source and target
            dimensionalities.
        hidden_channels (int): Size of each hidden sample.
        num_layers (int): Number of message passing layers.
        out_channels (int, optional): If not set to :obj:`None`, will apply a
            final linear transformation to convert hidden node embeddings to
            output size :obj:`out_channels`. (default: :obj:`None`)
        dropout (float, optional): Dropout probability. (default: :obj:`0.`)
        act (str or Callable, optional): The non-linear activation function to
            use. (default: :obj:`"relu"`)
        act_first (bool, optional): If set to :obj:`True`, activation is
            applied before normalization. (default: :obj:`False`)
        act_kwargs (Dict[str, Any], optional): Arguments passed to the
            respective activation function defined by :obj:`act`.
            (default: :obj:`None`)
        norm (str or Callable, optional): The normalization function to
            use. (default: :obj:`None`)
        norm_kwargs (Dict[str, Any], optional): Arguments passed to the
            respective normalization function defined by :obj:`norm`.
            (default: :obj:`None`)
        jk (str, optional): The Jumping Knowledge mode. If specified, the model
            will additionally apply a final linear transformation to transform
            node embeddings to the expected output feature dimensionality.
            (:obj:`None`, :obj:`"last"`, :obj:`"cat"`, :obj:`"max"`,
            :obj:`"lstm"`). (default: :obj:`None`)
        **kwargs (optional): Additional arguments of the underlying
            :class:`torch_geometric.nn.conv.MessagePassing` layers.
    """
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        num_layers: int,
        dropout: float = 0.0,
        dropout_input: bool = True,
        transform_input: bool = True,
        hidden_channels: Optional[int] = None,
        act: Optional[str] = "relu",
        act_kwargs: Optional[Dict[str, Any]] = None,
        act_first: bool = False,
        norm: Optional[str] = None,
        norm_kwargs: Optional[Dict[str, Any]] = None,
        jk: Optional[str] = None,
        **kwargs,
    ):
        super().__init__()
        if hidden_channels is None:
            hidden_channels = out_channels

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.hidden_channels = hidden_channels
        self.num_layers = num_layers

        self.dropout = dropout
        self.dropout_input = dropout_input
        self.act_first = act_first
        
        self.jk_mode = jk
        
        if transform_input:
            self.lin0 = nn.Linear(in_channels, hidden_channels)
            in_channels = hidden_channels
        
        if act is not None:
            self.act = activation_resolver(act, **(act_kwargs or {}))

        self.convs = nn.ModuleList()
        if num_layers > 1:
            self.convs.append(
                self.init_conv(in_channels, hidden_channels, **kwargs))
            in_channels = hidden_channels
            
        for _ in range(num_layers - 2):
            self.convs.append(
                self.init_conv(in_channels, hidden_channels, **kwargs))
            in_channels = hidden_channels
        
        if jk is None:
            self.convs.append(
                self.init_conv(in_channels, out_channels, **kwargs))
        else:
            self.convs.append(
                self.init_conv(in_channels, hidden_channels, **kwargs))
            
            if jk != "last":
                self.jk = JumpingKnowledge(jk, hidden_channels, num_layers)
        
            if jk == "cat":
                jk_channels = num_layers * hidden_channels
            else:
                jk_channels = hidden_channels
            self.lin_jk = nn.Linear(jk_channels, out_channels)
        
        if norm is not None:
            self.norms = nn.ModuleList()
            norm_layer = normalization_resolver(
                norm,
                hidden_channels,
                **(norm_kwargs or {}),
            )
            for _ in range(num_layers - 1):
                self.norms.append(copy.deepcopy(norm_layer))
            if jk is not None:
                self.norms.append(copy.deepcopy(norm_layer))
        
        self.reset_parameters()

    def init_conv(self,
        in_channels: int,
        out_channels: int,
        **kwargs
    ) -> nn.Module:
        raise NotImplementedError

    def reset_parameters(self):
        if hasattr(self, "lin0"):
            self.lin0.reset_parameters()
        
        if hasattr(self, "jk"):
            self.jk.reset_parameters()
            
        if hasattr(self, "lin_jk"):
            self.lin_jk.reset_parameters()
        
        for conv in self.convs:
            conv.reset_parameters()
        
        if hasattr(self, "norms"):
            for norm in self.norms:
                norm.reset_parameters()

    def forward(
        self,
        g: PGraph,
        x: Tensor,
        *,
        edge_attr: Optional[Tensor] = None,
    ) -> Tensor:
        if self.dropout_input:
            x = F.dropout(x, p=self.dropout, training=self.training)
        x = self.lin0(x) if hasattr(self, "lin0") else x
        
        xs: List[Tensor] = []
        for i in range(self.num_layers):
            argspec = inspect.getfullargspec(self.convs[i].forward)
            if "edge_attr" in argspec.args:
                x = self.convs[i](g, x, edge_attr=edge_attr)
            else:
                x = self.convs[i](g, x)
            
            x = F.dropout(x, p=self.dropout, training=self.training)
                
            if i == self.num_layers - 1 and self.jk_mode is None:
                break
            
            if hasattr(self, "act") and self.act_first:
                x = self.act(x)
            if hasattr(self, "norms"):
                x = self.norms[i](x)
            if hasattr(self, "act") and not self.act_first:
                x = self.act(x)
            
            # x = F.dropout(x, p=self.dropout, training=self.training)
                
            if hasattr(self, 'jk'):
                xs.append(x)

        x = self.jk(xs) if hasattr(self, 'jk') else x
        x = self.lin_jk(x) if hasattr(self, 'lin_jk') else x
        return x

    def __repr__(self) -> str:
        return (f'{self.__class__.__name__}({self.in_channels}, '
                f'{self.out_channels}, num_layers={self.num_layers})')