import torch
import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor
from typing import *

from core.distributed import PGraph
from torch_geometric.nn import Linear
import torch_geometric.nn.inits as pyg_inits



class GATConv(nn.Module):
    def __init__(self,
        in_channels: int,
        out_channels: int,
        heads: int = 1,
        concat: bool = True,
        negative_slope: float = 0.2,
        dropout: float = 0.0,
        edge_dim: Optional[int] = None,
        bias: bool = True,
    ):
        super().__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.heads = heads
        self.concat = concat
        self.negative_slope = negative_slope
        self.dropout = dropout
        self.edge_dim = edge_dim

        self.lin_node = Linear(in_channels, heads * out_channels, bias=False, weight_initializer="glorot")
        self.att_src = nn.Parameter(torch.Tensor(1, heads, out_channels))
        self.att_dst = nn.Parameter(torch.Tensor(1, heads, out_channels))

        if edge_dim is not None:
            self.lin_edge = nn.Linear(edge_dim, heads * out_channels, bias=False, weight_initializer="glorot")
            self.att_edge = nn.Parameter(torch.Tensor(1, heads, out_channels))
        else:
            self.register_parameter("lin_edge", None)
            self.register_parameter("att_edge", None)

        if bias and concat:
            self.bias = nn.Parameter(torch.Tensor(heads * out_channels))
        elif bias and not concat:
            self.bias = nn.Parameter(torch.Tensor(out_channels))
        else:
            self.register_parameter("bias", None)

        self.reset_parameters()

    def reset_parameters(self):
        self.lin_node.reset_parameters()
        if self.lin_edge is not None:
            self.lin_edge.reset_parameters()
        pyg_inits.glorot(self.att_src)
        pyg_inits.glorot(self.att_dst)
        pyg_inits.glorot(self.att_edge)
        pyg_inits.zeros(self.bias)

    def forward(self,
        g: PGraph,
        x: Tensor,
        *,
        edge_attr: Optional[Tensor] = None,
    ):
        x_dst, x_src = g.gather(x, src=True, dst=True)
        
        H, C = self.heads, self.out_channels
        x_src: Tensor = self.lin_node(x_src).view(-1, H, C)
        x_dst: Tensor = self.lin_node(x_dst).view(-1, H, C)
        
        alpha = self._compute_alpha(g, x_src, x_dst, edge_attr)
        w = g.softmax(alpha, src=False, dst=True)
        out = g.scatter(x_src * w, src=False, dst=True, reduce_op="sum")

        if self.concat:
            out = out.view(-1, self.heads * self.out_channels)
        else:
            out = out.mean(dim=1)

        if self.bias is not None:
            out += self.bias
            
        return out
    
    def _compute_alpha(self,
        g: PGraph,
        x_src: Tensor,
        x_dst: Tensor,
        edge_attr: Optional[Tensor],
    ) -> Tensor:
        H, C = self.heads, self.out_channels
        
        alpha_src = (x_src * self.att_src).sum(dim=-1, keepdim=True)
        alpha_dst = (x_dst * self.att_dst).sum(dim=-1, keepdim=True)
        alpha = alpha_src + alpha_dst
        
        if edge_attr is not None and self.lin_edge is not None:
            if edge_attr.dim() == 1:
                edge_attr = edge_attr.view(-1, 1)
            edge_attr = self.lin_edge(edge_attr).view(-1, H, C)
            alpha += (edge_attr * self.att_edge).sum(dim=-1, keepdim=True)
            
        alpha = F.leaky_relu(alpha, self.negative_slope)
        # alpha = alpha.exp()
        # alpha = alpha.sigmoid().exp()
        return alpha
    
    def _scatter(self,
        g: PGraph,
        x: Tensor,
        alpha: Tensor,
    ):
        s = 1.0 / g.scatter(alpha)
        # s[s.isinf()] = 0.0
        
        a = F.dropout(alpha, p=self.dropout, training=self.training)
        return g.scatter(x * a) * s

    def __repr__(self) -> str:
        return (f'{self.__class__.__name__}({self.in_channels}, '
                f'{self.out_channels}, heads={self.heads})')

    
    
