import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.distributed as dist
import torch.multiprocessing as mp

from torch import Tensor

import os
import os.path as osp

from typing import *

from core.distributed import PGraph
from model import GCN

from utils.functional import train_epoch, eval_epoch
from utils.partition import partition_load

import argparse

import time
import psutil

parser = argparse.ArgumentParser()
parser.add_argument("--data_path", type=str, required=True)
parser.add_argument("--part_algo", type=str, required=True)
parser.add_argument("--hidden_channels", type=int, default=64)
parser.add_argument("--num_layers", type=int, default=2)
parser.add_argument("--epochs", type=int, default=10)
parser.add_argument("--dropout", type=float, default=0.0)
parser.add_argument("--gpu", action="store")
parser.add_argument("--jk", type=str, default=None)
parser.add_argument("--norm", type=str, default=None)

def main(local_rank: int):
    args = parser.parse_args()
    rank = dist.get_rank()
    world_size = dist.get_world_size()
    
    if args.gpu and torch.cuda.is_available():
        device = torch.cuda.current_device()
    else:
        device = torch.device("cpu")
    
    data = partition_load(
        root=args.data_path,
        algo=args.part_algo,
    )
    
    g = PGraph(
        local_node_table=data.local_node_table,
        local_edge_table=data.local_edge_table,
        local_edge_route=data.local_edge_route,
        device=device
    ).enable_all_gather().enable_all_scatter()
    
    num_features = data.num_features
    num_classes = data.y.max().item() + 1
    print(f"{rank}: data loaded")
    
    net = GCN(
        in_channels=num_features,
        hidden_channels=args.hidden_channels,
        out_channels=num_classes,
        num_layers=args.num_layers,
        dropout=args.dropout,
        norm=args.norm,
        jk=args.jk,
    ).to(device)
    
    if device != torch.device("cpu"):
        net = nn.SyncBatchNorm.convert_sync_batchnorm(net)
    net = nn.parallel.DistributedDataParallel(net)
    
    opt = torch.optim.Adam(net.parameters(), lr=0.01, weight_decay=5e-4)
    
    print(f"{rank}: initialize model and optimizer")
    
    x = data.x.to(device)
    y = data.y.to(device)
    w = data.gcn_norm.to(device)
    train_mask = data.train_mask.to(device)
    val_mask = data.val_mask.to(device)
    test_mask = data.test_mask.to(device)
    
    avg_mem = 0.0
    avg_dur = 0.0
    avg_num = 0
    
    best_val_acc = best_test_acc = 0
    for ep in range(1, args.epochs+1):
        time_start = time.time()
        
        train_loss, train_acc = train_epoch(net, opt, g, x, y, w, train_mask)
        val_loss, val_acc = eval_epoch(net, g, x, y, w, val_mask)
        test_loss, test_acc = eval_epoch(net, g, x, y, w, test_mask)

        if val_acc > best_val_acc:
            best_val_acc = val_acc
            best_test_acc = test_acc
        
        duration = time.time() - time_start
        cur_mem = psutil.Process(os.getpid()).memory_info().rss
        cur_mem_mb = round(cur_mem / 1024**2)
        
        if ep > 1:
            avg_mem += cur_mem
            avg_dur += duration
            avg_num += 1
        if local_rank == 0:
            print(
                f"ep: {ep}, mem: {cur_mem_mb}MiB, duration: {duration:.2f}s, "
                f"loss: [{train_loss:.4f}/{val_loss:.4f}/{test_loss:.6f}], "
                f"accuracy: [{train_acc:.4f}/{val_acc:.4f}/{test_acc:.4f}], "
                f"best_accuracy: {best_test_acc:.4f}")
    
    avg_mem = round(avg_mem / avg_num / 1024**2)
    avg_dur = avg_dur / avg_num
    print(f"average memory: {avg_mem}MiB, average duration: {avg_dur:.2f}s")

if __name__ == "__main__":
    torch.set_num_threads(24)
    
    local_rank = int(os.environ.get("LOCAL_RANK", "0"))
    if torch.cuda.is_available():
        torch.cuda.set_device(local_rank)
    
    backend = os.environ.get("C10D_BACKEND", "gloo")
    dist.init_process_group(backend)
    
    main(local_rank)
        