import torch
import torch.distributed as dist

from torch import Tensor, LongTensor
from torch_sparse import SparseTensor

from torch_geometric.data import Data
from torch_geometric.utils import degree

import os
import os.path as osp
import shutil

from typing import *



def partition_save(root: str, data: Data, num_parts: int, algo: str = "dbh"):
    root = osp.abspath(root)
    if osp.exists(root) and not osp.isdir(root):
        raise ValueError(f"path '{root}' should be a directory")
    
    path = osp.join(root, f"{algo}_{num_parts}")
    if osp.exists(path) and not osp.isdir(path):
        raise ValueError(f"path '{path}' should be a directory")
    
    if osp.exists(path) and os.listdir(path):
        print(f"directory '{path}' not empty and cleared")
        for p in os.listdir(path):
            p = osp.join(path, p)
            if osp.isdir(p):
                shutil.rmtree(osp.join(path, p))
            else:
                os.remove(p)
                
    if not osp.exists(path):
        print(f"creating directory '{path}'")
        os.makedirs(path)
    
    pdata = partition_data(data, num_parts, algo, verbose=True)
    for i in range(num_parts):
        print(f"saving partition data: {i+1}/{num_parts}")
        fn = osp.join(path, f"{i:03d}")
        torch.save(pdata[i], fn)

def partition_load(root: str, algo: str = "dbh") -> Data:
    rank = dist.get_rank()
    world_size = dist.get_world_size()
    fn = osp.join(root, f"{algo}_{world_size}", f"{rank:03d}")
    return torch.load(fn)

def partition_data(data: Data, num_parts: int, algo: str = "dbh", verbose: bool = True) -> List[Data]:
    if algo == "metis":
        part_fn = metis
    elif algo == "dbh":
        part_fn = dbh
    else:
        raise ValueError(f"invalid algorithm: {algo}")
    
    num_nodes = data.num_nodes
    edge_index = data.edge_index
    
    if verbose: print(f"running partition algorithm: {algo}")
    node_parts, edge_parts = part_fn(edge_index, num_nodes, num_parts)
    
    if verbose: print("computing GCN normalized edge_attr")
    deg_j = degree(edge_index[0], num_nodes).pow(-0.5)
    deg_i = degree(edge_index[1], num_nodes).pow(-0.5)
    deg_i[deg_i.isinf() | deg_i.isnan()] = 0.0
    deg_j[deg_j.isinf() | deg_j.isnan()] = 0.0
    edge_attr = deg_j[data.edge_index[0]] * deg_i[data.edge_index[1]]
    
    npart_inds = []
    epart_inds = []
    
    if verbose: print("computing local partition table")
    pdata = [{} for _ in range(num_parts)]
    for i in range(num_parts):
        npart_i = torch.where(node_parts == i)[0]
        epart_i = torch.where(edge_parts == i)[0]
        npart_inds.append(npart_i)
        epart_inds.append(epart_i)
        
        npart = npart_i
        epart = edge_index[:,epart_i]
        
        route = epart.flatten().unique()
        route = torch.vstack([
            route, node_parts[route]
        ])
        gcn_norm = edge_attr[epart_i]
        
        pdata[i]["local_node_table"] = npart
        pdata[i]["local_edge_table"] = epart
        pdata[i]["local_edge_route"] = route
        pdata[i]["gcn_norm"] = gcn_norm
    
    if verbose: print("partition local features")
    for key, val in data:
        if key == "edge_index":
            continue
        if isinstance(val, Tensor):
            if val.size(0) == num_nodes:
                for i in range(num_parts):
                    npart_i = npart_inds[i]
                    pdata[i][key] = val[npart_i]
        elif isinstance(val, SparseTensor):
            pass
        else:
            for i in range(num_parts):
                pdata[i][key] = val
                
    for i in range(num_parts):
        pdata[i] = Data(**pdata[i])
    return pdata

def _nopart(edge_index: LongTensor, num_nodes: int) -> Tuple[LongTensor, LongTensor]:
    node_parts = torch.zeros(num_nodes, dtype=torch.long)
    edge_parts = torch.zeros(edge_index.size(1), dtype=torch.long)
    return node_parts, edge_parts

def metis(edge_index: LongTensor, num_nodes: int, num_parts: int) -> Tuple[LongTensor, LongTensor]:
    if num_parts <= 1:
        return _nopart(edge_index, num_nodes)
    
    adj_t = SparseTensor.from_edge_index(edge_index, sparse_sizes=(num_nodes, num_nodes)).to_symmetric()
    rowptr, col, _ = adj_t.csr()
    node_parts = torch.ops.torch_sparse.partition(rowptr, col, None, num_parts, num_parts < 8)
    edge_parts = node_parts[edge_index[1]]
    return node_parts, edge_parts


def dbh(edge_index: LongTensor, num_nodes: int, num_parts: int) -> Tuple[LongTensor, LongTensor]:
    if num_parts <= 1:
        return _nopart(edge_index, num_nodes)
    
    deg = degree(edge_index.flatten(), num_nodes, dtype=torch.long)
    deg_u = deg[edge_index[0]]
    deg_v = deg[edge_index[1]]
    
    node_parts = torch.arange(num_nodes).type_as(deg_u)
    edge_parts = torch.zeros_like(deg_v)
    
    select_u = deg_u <= deg_v
    select_v = deg_u >  deg_v
    edge_parts[select_u] = node_parts[edge_index[0]][select_u]
    edge_parts[select_v] = node_parts[edge_index[1]][select_v]
    
    return node_parts % num_parts, edge_parts % num_parts
